from chainer import reporter
from chainer.training import trigger as trigger_module
from chainer.training.extension import Extension

class SacredReport(Extension):
	def __init__(self, ex, keys=None, trigger=(1, "epoch")):
		super(SacredReport, self).__init__()
		self.ex = ex
		self._keys = keys
		self._trigger = trigger_module.get_trigger(trigger)

		self._init_summary()

	def initialize(self, trainer):
		pass

	def __call__(self, trainer):
		if self.ex is None:
			return

		obs = trainer.observation
		keys = self._keys

		if keys is None:
			self._summary.add(obs)
		else:
			self._summary.add({k: obs[k] for k in keys if k in obs})

		if not self._trigger(trainer):
			return

		stats = self._summary.compute_mean()
		epoch = trainer.updater.epoch
		for name in stats:
			self.ex.log_scalar(name, float(stats[name]), step=epoch)

		self._init_summary()

	def _init_summary(self):
		self._summary = reporter.DictSummary()



import chainer.functions as F

from chainer import variable
from chainer.initializers import Constant
from chainer_addons.functions import alpha_prod

from .compact_bilinear import CompactBilinearPooling

class AlphaPooling(CompactBilinearPooling):
	"""
		AlphaPooling implementation based on CompactBilinearPooling and
		Simon et al. (https://arxiv.org/abs/1705.00487)
	"""
	def __init__(self, init_alpha=1.0, eps=1e-5, use_built_ins=False, *args, **kwargs):
		super(AlphaPooling, self).__init__(*args, **kwargs)

		with self.init_scope():
			self.alpha = variable.Parameter(
				initializer=Constant(init_alpha), shape=(1,), name="alpha")
			self.add_persistent("eps", eps)

		self.use_built_ins = use_built_ins

	def __call__(self, x):

		if self.use_built_ins:
			sgn_x = F.sign(x)
			abs_x = F.absolute(x) + self.eps
			big_a = F.broadcast_to(self.alpha - 1, x.shape)
			x_hat = sgn_x * (abs_x ** big_a)
		else:
			x_hat = alpha_prod(x, self.alpha, eps=self.eps)

		res = super(AlphaPooling, self).__call__(x_hat, x)
		return res




import numpy as np
import chainer.links as L
import chainer.functions as F

from chainer.links.model.vision.vgg import prepare

from chainer_addons.models import PretrainedModelMixin

from collections import OrderedDict

class VGG19Layers(PretrainedModelMixin, L.VGG16Layers):
	"""
		adds some additional layers to the original VGG16
		model definition. the last linear layer is also replaced
		if number of classes is given. this is useful in case of
		fine-tuning
	"""

	class meta:
		classifier_layers = ["fc6", "fc7", "fc8"]
		conv_map_layer = "conv5_4"
		feature_layer = "fc7"
		feature_size = 4096
		n_conv_maps = 512
		input_size = 224
		mean = np.array([103.939, 116.779, 123.68], dtype=np.float32).reshape(3,1,1)
		prepare_func = prepare

	@staticmethod
	def pooling(x):
		return F.max_pooling_2d(x, ksize=2, stride=2)

	def init_layers(self, n_classes, **kwargs):
		self.conv3_4 = L.Convolution2D(256, 256, 3, 1, 1)
		self.conv4_4 = L.Convolution2D(512, 512, 3, 1, 1)
		self.conv5_4 = L.Convolution2D(512, 512, 3, 1, 1)


	@property
	def _links(self):
		links = [
			('conv1_1',  [self.conv1_1, F.relu]),
			('conv1_2',  [self.conv1_2, F.relu]),
			('pool1',    [VGG19Layers.pooling]),
			('conv2_1',  [self.conv2_1, F.relu]),
			('conv2_2',  [self.conv2_2, F.relu]),
			('pool2',    [VGG19Layers.pooling]),
			('conv3_1',  [self.conv3_1, F.relu]),
			('conv3_2',  [self.conv3_2, F.relu]),
			('conv3_3',  [self.conv3_3, F.relu]),
			('conv3_4',  [self.conv3_4, F.relu]),
			('pool3',    [VGG19Layers.pooling]),
			('conv4_1',  [self.conv4_1, F.relu]),
			('conv4_2',  [self.conv4_2, F.relu]),
			('conv4_3',  [self.conv4_3, F.relu]),
			('conv4_4',  [self.conv4_4, F.relu]),
			('pool4',    [VGG19Layers.pooling]),
			('conv5_1',  [self.conv5_1, F.relu]),
			('conv5_2',  [self.conv5_2, F.relu]),
			('conv5_3',  [self.conv5_3, F.relu]),
			('conv5_4',  [self.conv5_4, F.relu]),
			('pool5',    [VGG19Layers.pooling]),
		]

		if all([hasattr(self, attr) for attr in ["fc6", "fc7", "fc8"]]):
			links += [
				('fc6',      [self.fc6, F.relu, F.dropout]),
				('fc7',      [self.fc7, F.relu, F.dropout]),
				('fc8',      [self.fc8]),
				('prob',     [F.softmax]),
			]
		return links



from .compact_bilinear import CompactBilinearPooling
from .alpha_pooling import AlphaPooling
from .global_avg import GlobalAveragePooling
from chainer_addons.functions import tf_average_pooling_2d

from cvargparse.utils.enumerations import BaseChoiceType
import chainer.functions as F
from functools import partial

def pool_creator(func):
	def creator(ksize, stride=None, pad=0, **kwargs):
		return partial(func, ksize=ksize, stride=stride, pad=pad)
	return partial(creator)

class PoolingType(BaseChoiceType):
	MAX = pool_creator(F.max_pooling_2d)
	AVG = pool_creator(F.average_pooling_2d)
	TF_AVG = pool_creator(tf_average_pooling_2d)

	G_AVG = GlobalAveragePooling
	CBIL = CompactBilinearPooling
	ALPHA = AlphaPooling

	Default = G_AVG

	@classmethod
	def new(cls, pool_type, **kwargs):
		pool_cls = cls.get(pool_type).value
		return pool_cls(**kwargs)




import numpy as np
from functools import reduce, partial, wraps

def asarray(func):

	@wraps(func)
	def inner(image, *args, **kw):
		from PIL import Image
		from scipy.misc import fromimage, toimage
		is_pil = isinstance(image, Image.Image)
		if is_pil:
			image = fromimage(image)
		image = func(image, *args, **kw)
		if is_pil:
			image = toimage(image)
		return image

	return inner

def crop(image, x,y,w,h):
	return image[:, y: y + h, x: x + w]

@asarray
def random_horizontal_flip(image, threshold=.5, axis=2):
	if np.random.random() < threshold:
		return np.flip(image, axis=axis)
	return image

@asarray
def random_crop(image, crop_size):
	th, tw = crop_size
	c, h, w = image.shape
	th, tw = min(th, h), min(tw, w)
	x = np.random.randint(0, w - tw)
	y = np.random.randint(0, h - th)
	return crop(image, x, y, th, tw)

@asarray
def center_crop(image, crop_size):
	th, tw = crop_size
	c, h, w = image.shape
	th, tw = min(th, h), min(tw, w)
	x = int((w - tw) / 2.)
	y = int((h - th) / 2.)
	return crop(image, x, y, th, tw)


class Augmentation(object):
	def __init__(self):
		self.augmentations = []

	def __call__(self, image):
		if not self.augmentations: return image
		def call(val, func):
			return func(val)
		return reduce(
			call, self.augmentations,
			image)

	def center_crop(self, size):
		self.augmentations.append(
			partial(center_crop, crop_size=size))
		return self

	def random_crop(self, size):
		self.augmentations.append(
			partial(random_crop, crop_size=size))
		return self

	def random_horizontal_flip(self, threshold=.5, axis=2):

		self.augmentations.append(
			partial(random_horizontal_flip, threshold=threshold, axis=axis))
		return self



import numpy as np, chainer
from chainer.training import extension

class ContiniousLearningRate(extension.Extension):
	trigger = 1, 'epoch'

	def _get_optimizer(self, trainer):
		return self._optimizer or trainer.updater.get_optimizer('main')

	def initialize(self, trainer):
		self.set_lr(trainer, self._lr)

	def set_lr(self, trainer, value):
		optimizer = self._get_optimizer(trainer)
		setattr(optimizer, self._attr, max(value, self._target))

	def __init__(self, attr, lr, target, epochs, offset, optimizer=None):
		super(ContiniousLearningRate, self).__init__()
		self._optimizer = optimizer
		self._attr = attr
		self._lr = lr
		self._target = target
		self._epochs = epochs
		self._offset = offset

	@property
	def factor(self):
		raise NotImplementedError

	def new_lr(self, epoch):
		raise NotImplementedError

	def __call__(self, trainer):
		if trainer.updater.epoch - self._offset < 0: return
		if trainer.updater.epoch - self._offset > self._epochs: return
		self.set_lr(trainer, self.new_lr(trainer.updater.epoch - self._offset))


class ExponentialLearningRate(ContiniousLearningRate):

	@property
	def factor(self):
		return pow(10, np.log10(self._target / self._lr) / self._epochs)

	def new_lr(self, epoch):
		return self._lr * self.factor ** epoch

class LinearLearningRate(ContiniousLearningRate):

	def new_lr(self, epoch):
		return self._lr - (self.factor * epoch)

	@property
	def factor(self):
		return (self._lr - self._target) / self._epochs



from __future__ import division
import datetime, sys, time

from chainer.training import extension
from chainer.training.extensions import util
from chainer.training import trigger


class JupyterProgressBar(extension.Extension):

	def __init__(self, training_length=None, update_interval=100,
				 bar_length=50):
		self._training_length = training_length
		self._status_template = None
		self._update_interval = update_interval
		self._bar_length = bar_length
		self._recent_timing = []

		self.progress = None
		self.start = None
		if self._training_length is not None:
			self.init_bar()


	def init_bar(self):
		from ipywidgets import IntProgress, HTML, VBox
		from IPython.display import display
		size = self._training_length[0]
		self.total_progress = IntProgress(min=0, max=self._bar_length, value=0)
		self.epoch_progress = IntProgress(min=0, max=self._bar_length, value=0)

		self.total_progress.bar_style = 'info'
		self.epoch_progress.bar_style = 'info'

		self.progress_label = HTML()
		self.eta_label = HTML()

		box = VBox(children=[
			self.eta_label,
			self.total_progress,
			self.epoch_progress,
			self.progress_label,
		])
		display(box)


	def __call__(self, trainer):
		# initialize some attributes at the first call
		if self._training_length is None:
			t = trainer.stop_trigger
			if not isinstance(t, trigger.IntervalTrigger):
				raise TypeError(
					'cannot retrieve the training length from %s' % type(t))
			self._training_length = t.period, t.unit
			self.init_bar()

		if self._status_template is None:
			self._status_template = " | ".join([
				'{0.iteration:10} iter',
				'{0.epoch} epoch / %s %ss' % self._training_length,
				'Training time: {1}'
			])

		length, unit = self._training_length

		iteration = trainer.updater.iteration

		# print the progress bar
		if iteration % self._update_interval != 0: return

		epoch = trainer.updater.epoch_detail
		self._recent_timing = self._recent_timing
		now = time.time()

		self._recent_timing.append((iteration, epoch, now))

		if unit == 'iteration':
			rate = iteration / length
		else:
			rate = epoch / length
		epoch_rate = epoch - int(epoch)

		self.total_progress.value = int(rate * self._bar_length)
		self.epoch_progress.value = int(epoch_rate * self._bar_length)
		self.progress_label.value = " | ".join([
			"This epoch: {: 6.2%}".format(epoch_rate),
			"total: {: 6.2%}".format(rate),
		])

		old_t, old_e, old_sec = self._recent_timing[0]
		span = now - old_sec

		if span != 0:
			speed_t = (iteration - old_t) / span
			speed_e = (epoch - old_e) / span
		else:
			speed_t = float('inf')
			speed_e = float('inf')

		if unit == 'iteration':
			ETA_total = (length - iteration) / speed_t
		else:
			ETA_total = (length - epoch) / speed_e

		ETA_epoch = (1 - epoch_rate) / speed_e

		self.eta_label.value = "<br>".join([
			self._status_template.format(
				trainer.updater,
				0 if self.start is None else datetime.timedelta(seconds=now - self.start)
			),
			"Time per iter: {} and epoch: {}".format(
				datetime.timedelta(seconds=1 / speed_t),
				datetime.timedelta(seconds=int(1/speed_e))
			),
			"Epoch ready in: {} and training in: {}".format(
				datetime.timedelta(seconds=int(ETA_epoch)),
				datetime.timedelta(seconds=int(ETA_total))
			)
		])


		if len(self._recent_timing) > 100:
			self._recent_timing.pop(0)

	def initialize(self, trainer):
		self.start = time.time()

	def finalize(self):
		if hasattr(self, "total_progress"):
			self.total_progress.bar_style = 'success'
		if hasattr(self, "epoch_progress"):
			self.epoch_progress.bar_style = 'success'



import chainer
import chainer.links as L
import chainer.functions as F

import numpy as np
import six
from os.path import isfile
from skimage.transform import resize

from chainer_addons.models import PretrainedModelMixin

from functools import partial


class AlexNet(PretrainedModelMixin, chainer.Chain):

	class meta:
		classifier_layers = ["fc6", "fc7", "fc8"]
		conv_map_layer = "pool5"
		feature_layer = "fc7"
		feature_size = 4096
		n_conv_maps = 256
		input_size = 227

		mean = None

		@classmethod
		def get_mean(cls, f='/tmp/ilsvrc_2012_mean.npy'):
			if not isfile(f) and cls.mean is None:
				print('Downloading ILSVRC12 mean file for NumPy...')
				six.moves.urllib.request.urlretrieve(
					'https://github.com/BVLC/caffe/raw/master/python/caffe/imagenet/ilsvrc_2012_mean.npy', f)
				print('Done')

			print('Loading ILSVRC12 mean file from \"{}\"'.format(f))
			cls.mean = np.load(f).astype(np.float32)
			c,h,w = cls.mean.shape

			dh, dw = int((h - cls.input_size) / 2), int((w - cls.input_size) // 2)
			cls.mean = cls.mean[:, dh:dh + cls.input_size, dw:dw + cls.input_size]

		@staticmethod
		def prepare_func(x, size=None):
			x = resize(x, size, mode="constant") * 255
			x = x.astype(np.float32).transpose(2,0,1)
			x = x[::-1]
			return x - AlexNet.meta.mean

	@classmethod
	def prepare_back(cls, img):
		img = (img + cls.meta.mean).transpose(1,2,0)
		img = img[..., ::-1].astype(np.uint8)
		return img

	@property
	def _links(self):

		links = [
			("conv1", [
				self.conv1, F.relu,
				partial(F.max_pooling_2d, ksize=3, stride=2),
				F.local_response_normalization]
			),
			("conv2", [
				self.conv2, F.relu,
				partial(F.max_pooling_2d, ksize=3, stride=2),
				F.local_response_normalization]
			),
			("conv3", [self.conv3, F.relu]),
			("conv4", [self.conv4, F.relu]),
			("conv5", [self.conv5, F.relu]),
			("pool5", [partial(F.max_pooling_2d, ksize=3, stride=2)]),
			('fc6', [self.fc6, F.relu]),
			('fc7', [self.fc7, F.relu]),
			('fc8', [self.fc8]),
			('prob', [F.softmax]),
		]

		return links



	def init_layers(self, n_classes, **kwargs):
		AlexNet.meta.get_mean()

		self.conv1 = L.Convolution2D(  3,  96, 11, stride=4, pad=0)
		self.conv2 = L.Convolution2D( 96, 256,  5, stride=1, pad=2)
		self.conv3 = L.Convolution2D(256, 384,  3, stride=1, pad=1)
		self.conv4 = L.Convolution2D(384, 384,  3, stride=1, pad=1)
		self.conv5 = L.Convolution2D(384, 256,  3, stride=1, pad=1)

		self.fc6 = L.Linear(13*13*256, 4096)
		self.fc7 = L.Linear(4096, 4096)
		self.fc8 = L.Linear(4096, n_classes)


	def __call__(self, x, layer_name='fc'):
		for key, funcs in self.functions.items():
			for func in funcs:
				x = func(x)
			if key == layer_name:
				return x



from chainer.function import Function
from chainer import cuda

import numpy as np

class AlphaProd(Function):
	"""
		Implementation as it is in the Paper:
			with an epsilon for numerical stability
	"""
	def __init__(self, eps):
		super(AlphaProd, self).__init__()
		self.eps = eps

	def forward(self, inputs):
		xp = cuda.get_array_module(*inputs)
		x, alpha = inputs

		abs_x = xp.abs(x) + self.eps
		return xp.sign(x) * xp.power(abs_x, alpha - 1),

	def backward_cpu(self, inputs, gys):
		xp = cuda.get_array_module(*inputs)
		x, alpha = inputs
		gy = gys[0]

		abs_x = xp.abs(x) + self.eps
		sgn_x = xp.sign(x)

		gx = xp.power(abs_x, alpha - 2) * xp.power(sgn_x, 2) * (alpha - 1)
		ga = xp.power(abs_x, alpha - 1) * sgn_x * xp.log(abs_x)

		gx = gx*gy
		ga = (ga*gy).sum().reshape(1)

		return gx, ga

	def backward_gpu(self, inputs, gys):
		x, alpha = inputs
		xp = cuda.cupy

		gx = xp.zeros_like(gys[0])
		ga = xp.zeros_like(gys[0])

		cuda.elementwise(
			in_params="B mask, T abs_x, T sgn_x, T alpha, T gy",
			out_params="T gx, T ga",
			operation="""
				gx = mask ? 0 : gy * pow(abs_x, alpha - 2) * (alpha - 1) ;
				ga = mask ? 0 : gy * pow(abs_x, alpha - 1) * sgn_x * log(abs_x) ;
			""",
			name="alpha_prod_bwd")(
				x==0, xp.abs(x) + self.eps, xp.sign(x), alpha, gys[0],
				gx, ga)

		return gx, ga.sum().reshape(1)


class AlphaProd2(Function):
	"""
		Implementation with the assumption that all inputs are positive

		as the consequence following holds:
		- x == abs(x)
		- sign(x) == sign(x)**2
		- the gradient for all x == 0 is 0 as well
		- sign(x) * <anything> == <anything> for all x != 0

		hence we only compute gradient for x != 0
	"""
	def forward(self, inputs):
		xp = cuda.get_array_module(*inputs)
		x, alpha = inputs
		return xp.power(x, alpha - 1),

	def backward_cpu(self, inputs, gys):
		x, alpha = inputs
		gy = gys[0]

		mask = x != 0
		gx = np.zeros_like(gy)
		ga = np.zeros_like(gy)

		if mask.any():
			gx[mask] = np.power(x[mask], alpha - 2) * (alpha - 1)
			ga[mask] = np.power(x[mask], alpha - 1) * np.log(x[mask])

		gx = gx*gy
		ga = (ga*gy).sum().reshape(1)
		return gx, ga

	def backward_gpu(self, inputs, gys):
		x, alpha = inputs

		gx = cuda.cupy.zeros_like(gys[0])
		ga = cuda.cupy.zeros_like(gys[0])

		cuda.elementwise(
			in_params="B mask, T x, T alpha, T gy",
			out_params="T gx, T ga",
			operation="""
				gx = mask ? 0 : gy * pow(x, alpha - 2) * (alpha - 1) ;
				ga = mask ? 0 : gy * pow(x, alpha - 1) * log(x) ;
			""",
			name="alpha_prod2_bwd")(
				x==0, x, alpha, gys[0],
				gx, ga)

		return gx, ga.sum().reshape(1)


def alpha_prod(x, alpha, eps=1e-5):
	# return AlphaProd(eps)(x, alpha)
	return AlphaProd2()(x, alpha)
